/*
-- This is        kludge_CLI.h

-- the command line interface, presented to clients
*/

#include <iostream>
#include <string>
#include <vector>    //provides the vector<> template
#include "kludge_engine.h"





#ifndef _KLUDGE_CLI_H
#define _KLUDGE_CLI_H

#define MAX_CLI_INPUT_LINE_LEN 200 
/* the buffer used is actually always MAX_LEN + 1 bytes long. */



typedef struct {
             size_t end_pos;     // end position of this token in the input line
             string token_data;  // the data content of the token
             char punct_type;    // for straight `switch-case' resolution
             bool punctuation;   // if false, this token has some data content
               } Kludge_Token;            // if true, it's a punctuation token



class Kludge_Path
{
public:
   bool leading_slash;      //if false, it's a relative path
   bool trailing_slash;     //if true, it's a directory-style container
   vector<string*> elements;
   Kludge_Path();
   ~Kludge_Path();
};



// This should become a vector of pointers
// and the tokenizer class should provide a method to clear the chain.
// Currently there's quite some copy construction traffic going on.
typedef vector<Kludge_Token> Kludge_Token_Chain;



enum Kludge_CLI_CMD_PARM_Type
{
// PARM_UNDEF,      // this should never occur
 PARM_PATH,       // class, instance or property DN
 PARM_CLASS,      // class (name [, contents])
 PARM_INSTANCE,   // instance (name, PARM_OBJ_CLASS class, PARM_OBJ_CLSREL cls_rel [, contents])
 PARM_CLSPROP,    // class property (PARM_NAME, PARM_VALUE)
 PARM_CLS_BLOB,   // a CLASS BLOB - one of CLS_BLOB, CLS_TXT_BLOB or METHOD
 PARM_INST_BLOB,  // an INSTANCE BLOB - one of INST_BLOB or INST_TXT_BLOB
 PARM_INSTPROP,   // instance property (content=name, class, value)
 PARM_CLSREL,     // class relationship ({PARM_VARIABLE}* : Master, Slave, Name_M, Name_S are mandatory)
 PARM_INSTREL,    // instance relationship (PARM_OBJ_CLASS class, PARM_OBJ_CLSREL cls_rel, {PARM_VARIABLE}* : Master, Slave, Name_M, Name_S are mandatory)
 PARM_NAME,       // property name attribute
 PARM_VALUE,      // property value attribute
 PARM_BLOB_VALUE, // blob value attribute (distinguished from plain property value to prevent mishandling/overruns)
 PARM_OBJ_CLASS,  // instance or instance_property class or cls_rel attribute
 PARM_OBJ_CLSREL, // instance or instance_property class or cls_rel attribute
 PARM_VARIABLE,   // general-purpose variable container (PARM_NAME, PARM_VALUE)
 PARM_ARGLIST,    // a set of variables - named method arguments
 PARM_OPTION      // a modifier option - currently the only option known is RECURSIVE with DELETE
 //PARM_EXPR        // arithmetical/logical expression ? Dunno :-(
};



class Kludge_CLI_CMD_PARM
{
public:
   Kludge_CLI_CMD_PARM(Kludge_CLI_CMD_PARM_Type parm_type);
   Kludge_CLI_CMD_PARM(Kludge_CLI_CMD_PARM_Type parm_type, 
                       string& cont) : type(parm_type), content(cont) {};
   Kludge_CLI_CMD_PARM(Kludge_CLI_CMD_PARM_Type parm_type, 
                       char* cont) : type(parm_type), content(cont) {};
   ~Kludge_CLI_CMD_PARM();
   int add_content(Kludge_CLI_CMD_PARM* cont_to_add);
   Kludge_CLI_CMD_PARM_Type type;
   string content;
   vector<Kludge_CLI_CMD_PARM*> contents;
};



enum Kludge_CLI_Command_Type
{
 CMD_GET,
 CMD_CD,
 CMD_EXIT,
 CMD_SHUTDOWN,
 CMD_PWD,
 CMD_INSERT,
 CMD_DELETE,
 CMD_SET,
 CMD_MOVE,
 CMD_RENAME,
 CMD_RUN
};



class Kludge_CLI_Command
{
public:
   Kludge_CLI_Command(Kludge_CLI_Command_Type cmd_type) : type(cmd_type) {};
   ~Kludge_CLI_Command();
   Kludge_CLI_Command_Type type;
   int add_param(Kludge_CLI_CMD_PARM* param);
   vector<Kludge_CLI_CMD_PARM*> params;
};



class Kludge_CLI; // for a mutually recursive reference

class Kludge_Tokenizer
{
// The tokenizer does not convert backslashed escape sequences - for now,
// I leave this up to higher-level functions working on top of the tokens.
// The leading backslash is registered, but the sequence passes through
// the tokenizer unprocessed.
public:
   Kludge_Tokenizer(iostream* strmin, iostream* strmout = NULL);
   ~Kludge_Tokenizer();
   Kludge_Token* Pop_Token();
   Kludge_Token* Peek_Token();
   int Reset();
private:
   iostream*     sin;            // an input stream of whatever sort
   iostream*     sout;           // an output stream for error reporting
   Kludge_Token_Chain  line_chain;     // a chain of tokens recovered from one line of input
   size_t              line_pos;       // position in the token chain
   char*               CLI_input_line; // a buffer for one line of input
   bool done;
   int Get_Line();
   Kludge_Token* Get_Token(string* line_parsed, size_t begin_here);
   size_t Get_Delimiters(string* line_parsed, size_t begin_here);
   bool Tokenize_Line();
   friend class Kludge_CLI; // mainly for Parse_Blob_Def() - needs direct access to the line buffer
   friend void Sig_Quit_Handler(int sig_num);
};



class Kludge_Thread;

class Kludge_CLI {
public:
   Kludge_CLI(iostream* input_str, iostream* output_str, Kludge_Thread* thread);
   ~Kludge_CLI();
   int Run();
   int Indent();
   int Unindent();
   static Kludge_pl_params* Prepare_Method_Params(Kludge_CLI_CMD_PARM* params);
   iostream*            input_stream;    // We need two streams even for a single
   iostream*            output_stream;   // underlying socket :(
   string               ind_spaces;
   size_t               ind_depth;
   bool                 indent;
   Kludge_Object*       pwd;
   Kludge_Thread*       thr;             // When it comes to who owns whom,
                                         // know ye, that the thread owns the CLI

   friend void Sig_Quit_Handler(int sig_num);
private:
   bool                 CLI_done;        // Are we finished with this client?
   Kludge_Tokenizer*    tokenizer;
   int Welcome_Message();
   Kludge_CLI_Command* Prepare_Command();
   int Execute_Command(Kludge_CLI_Command* cmd);
   Kludge_CLI_CMD_PARM* Parse_Var_Def();
   Kludge_CLI_CMD_PARM* Parse_Blob_Def();
   Kludge_CLI_CMD_PARM* Parse_Obj_Def();
   Kludge_CLI_CMD_PARM* Parse_Arg_List(bool check_semicolon = true);
   int Process_Command();
};



// ### internal kludge_CLI global functions

Kludge_Path* Kludge_CLI_String2path(string *str_parsed, iostream* error_output);

// In the following function, setting `base' to NULL or leaving it undefined
// effectively means `base = Kludge_Root`
Kludge_Object* Kludge_CLI_Path2Ptr(Kludge_Path* path, Kludge_CLI* CLI,
                Kludge_Object* base = NULL, bool skip_last=false, size_t level=0);

//string* Kludge_CLI_Path2string(Kludge_Path *path_to_combine, iostream* error_output);
//Kludge_Path* Kludge_CLI_PWD(Kludge_Object* ptr, iostream* error_output);
//
// I can't think of any use for a Kludge_Path, derived directly from a known Kludge_Object*.
// OTOH, no doubt I need to be able to derive the combined root-based path string from an
// existing known Kludge_Object* - to report it back to the client at the CLI. Human clients
// shall not deal with OID's, let alone server-side Kludge_Object pointers. Therefore, I'll
// frequently need to perform the `pwd' operation - show the path to a particular
// Kludge_Object. For this purpose, the generation of a Kludge_Path might prove cumbersome
// - it would introduce unnecessary overhead, consisting of alloc/create/destroy/dealloc 
// of a handful of string obects each time a path is derived. While in fact all we need is
// a single string to accumulate the whole path.
//
string* Kludge_CLI_PWD(Kludge_Object* ptr, iostream* error_output);

int Bruce_Eckel_stringCmpi(const string& s1, const string& s2);

#ifndef _KLUDGE_THREADS_H
#include "kludge_threads.h"
#endif

#endif // _KLUDGE_CLI_H
