
/*

This is          kludge_engine.h

-- This module provides the heart of Kludge: --
--       the object-oriented metamodel.      --



There are six basic Kludge object types:

* Kludge_Class
* Kludge_Class_Property
* Kludge_Class_Relationship
* Kludge_Instance
* Kludge_Instance_Property
* Kludge_Instance_Relationship

and there's a common abstract base class, the Kludge_Object.
And, an artificial common root for the two hierarchies (Kludge
classes and Kludge instances) - the Kludge_Root_Object.

The two "property" types also serve as hooks where BLOB types
are anchored (including METHOD, which is more or less a text blob).
After all, the BLOB types are leaf data objects just like the basic
properties. The three essential BLOB types are binary blobs,
text blobs and methods - the latter two sharing the same relational
table.
The "relationship" types have helper stubs to hook them up into
the Kludge_Object family inheritance (C++ inheritance). The stubs
are heirs of Kludge_Object - the relationships themselves are not.

The branching types, i.e. the Kludge_Class and Kludge_Instance, are
using libglib's GTree type for various ordered lists (indexes),
both at global and relative level.
The GTree is a plain C animal - it uses plain C (global syntax)
callback functions for comparisons and traversals. Interestingly,
in the GNU C++, static member functions are equal to global
standalone functions (after some implicit name decorations done
by the preprocessor/compiler) - which allows us to encapsulate
relevant traversal functions into the respective classes and even
make use of some of the OO syntactic candy to save typing.

Historically, all the traversal functions were global (standalone)
- it was only during the process of me having C++ dawn on me that
I memberized most of them.
Some of the traversal methods could probably be shared among the
different Kludge object types using C++ inheritance/polymorphism
via the common abstract base class called Kludge_Object. It's just
that I'm not completely sure that for the future this makes sense
and whether or not performance could suffer from the late binding
implied... or is there no such implication?

*/


// The following two macros are defined both by standard C++ #includes and by Perl stuff.
#undef Move
#undef assert
#include <string>  //provides the string class
#include <fstream> //provides file streams (e.g., cout)
#include <strstream> //provides rudimentary "string" streams - for SQL ops
#include <glib.h>  //provides GTrees
#include <vector>  //provides the vector<> template
#include "kludge_mysql.h"
#include "kludge_perl.h"
#include "base64.h"
#include "pcre.h"
using namespace std;



#ifndef _KLUDGE_ENGINE_H
#define _KLUDGE_ENGINE_H


typedef unsigned int OID_Type;

/* ### classes ### */


// to simplify mutually-recursive class declaration & #inclusion :
class Kludge_CLI;
class Kludge_CLI_CMD_PARM;

class Kludge_Object
{
public:
   enum Kludge_Object_Type    // For RTTI in Kludge_CLI::Execute_Command() 
   {                          // to differentiate between pointers returned by Kludge_CLI_Path2ptr()
      KLUDGE_ROOT,
      KLUDGE_CLASS,
      KLUDGE_CLASS_PROPERTY,
      //KLUDGE_CLASS_RELATIONSHIP,
      KLUDGE_CLASS_REL_STUB,
      KLUDGE_INSTANCE,
      KLUDGE_INSTANCE_PROPERTY,
      //KLUDGE_INSTANCE_RELATIONSHIP,
      KLUDGE_INST_REL_STUB
   };
   typedef struct    // when deleting children or properties
   {
      Kludge_CLI* CLI;
      vector<Kludge_Object*> objs_deleted;
   } del_param;
   
   Kludge_Object_Type Type;   // Cannot be set within the ctor initializer list in the child classes.

   virtual ~Kludge_Object() = 0;
   virtual int List_Contents(Kludge_CLI* CLI, bool recursive = false) = 0;
   virtual Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false) = 0;
   virtual int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI) = 0;
   virtual int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI) = 0; // only makes sense with properties
   virtual int Move(Kludge_Object* destination, Kludge_CLI* CLI) = 0;
   virtual int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI) = 0;
   virtual int Delete(Kludge_CLI* CLI, bool recursive = false) = 0;
   virtual int DB_Delete(Kludge_CLI* CLI, bool recursive = false) = 0;
   virtual int Zap() = 0; // only works in RAM, not DB
   virtual int pwd(string* result) = 0;
   virtual string* My_Name() = 0; // Just an initial toy for Perl - delete this when not useful anymore
   virtual Kludge_Object* Superior() = 0;
   
   // careful with the operators - boolean/linear algebra is not a complete analog.
   // Our space is a tree - not a numeric linear set.
   // A > B means A is superior to B in the tree. 
   // Which *doesn't* imply that e.g.     !(A >= B) ==> A < B 
   //  or that                            (!(A > B)  &&  !(A < B)) ==> A == B
   // In fact, the following *is* perfectly possible:   !(A < B) && !(A > B) && !(A == B)
   // If neither object is superior to the other one,
   //  they may simply reside at different branches of the tree.
   int operator>(const Kludge_Object& rhs) const;
   int operator>=(const Kludge_Object& rhs) const;
   int operator<(const Kludge_Object& rhs) const;
   int operator<=(const Kludge_Object& rhs) const;
};



// for a mutually recursive declaration :
class Kludge_Class_Property;
class Kludge_Class_Rel_Stub;

// Corresponding to an entry from table Classes
class Kludge_Class : public Kludge_Object
{
public:
   OID_Type Class;
   string Name;
   Kludge_Class *Parent;      // base class - class to inherit from
   // The following two static members point to global indexes of Kludge_Class objects.
   // Both are initialized at the top of kludge_engine.cc (mandatory static member init).
   static GTree* By_ID;
   static GTree* By_Name;
   GTree* Children;           // derived classes - classes that inherit from us
   GTree* Properties;
   GTree* Inherited_Props;
   GTree* Relationships;      // if NULL, relationships index not initialized
   GTree* Inherited_Rels;
   GTree* Instances;          // for instances and instance properties
   bool ctor_failed;
   bool relationships_loaded; // if false, relationships index not authoritative (not necessarily complete) 
   bool inheritance_done;
   // bool children_loaded;  // Would only be any use with partial loading.

   // when inserting a new class:
   Kludge_Class(string* class_name, Kludge_Class* parent_class, Kludge_CLI* CLI);
   // when loading a class by global name (nowadays only the root class):
   Kludge_Class(const char *class_name, Kludge_Class* parent_class, Kludge_CLI* CLI);
   // when loading a class by ID (do we use this at all?):
   Kludge_Class(OID_Type ClassID, Kludge_Class* parent_class, Kludge_CLI* CLI);
   // when loading a flurry of classes (e.g., children) based on a single SQL query (to save CPU time):
   Kludge_Class(MYSQL_ROW DB_record, Kludge_Class* parent_class, Kludge_CLI* CLI);
   ~Kludge_Class();
   int Load_Children(Kludge_CLI* CLI, bool recursive = false); // DB access function
   bool Free_Children();
   int Load_Properties(Kludge_CLI* CLI, bool initiate_inheritance = true); // DB access function
   bool Free_Properties();
   int Init_Relationships();
   int Load_Relationships(Kludge_CLI* CLI, bool initiate_inheritance = true); // DB access function
   bool Free_Relationships();
   int Init_Instances();
   bool Free_Instances();
   int Inherit(Kludge_CLI* CLI);    // Build/rebuild your inheritance indexes - now. Lookup required.
   int ReInherit(Kludge_CLI* CLI);  // Something has changed in your ancestry - rebuild your inheritance index, if you've already built it. If not, don't bother.
   int ReInherit_Children(Kludge_CLI* CLI);
   int UnInherit();  // Something has changed in your ancestry. Invalidate your inheritance index, if you've already built it - and tell your children. If not, don't bother.
   int UnInherit_Children();
   // In the following two function headers, "NULL" actually means "Parent". (Static declaration workaround.)
   int Add_To_Index(Kludge_Class* Where = NULL);
   int Free_From_Index(Kludge_Class* Where = NULL); // local only - only the "child" entry
   bool Free_From_Global_Index(); // Does not delete the "child" entry from the Parent
   int List_Properties(Kludge_CLI* CLI, bool recursive = false); 
   int List_Children(Kludge_CLI* CLI, bool recursive = false);
   int List_Relationships(Kludge_CLI* CLI, bool recursive = false);
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int Describe_Self(Kludge_CLI* CLI);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   Kludge_Object* Resolve_Content_Name_Enh(string* Content_Name, Kludge_CLI* CLI,
                                           bool own_only = false, bool initiate_inheritance = true);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI); // DB access function; does not delete obj_to_insert
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int Delete(Kludge_CLI* CLI, bool recursive = false);    // indirect DB access function
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false); // DB access function
   int Zap();                               // memory only cleanup
   Kludge_Class* Lookup_Child(string* name_to_lookup, Kludge_CLI* CLI);
   Kludge_Class_Property* Lookup_Property(string* name_to_lookup, Kludge_CLI* CLI,
                            bool own_only = false, bool initiate_inheritance = true);
   Kludge_Class_Rel_Stub* Lookup_Relationship(string* name_to_lookup, Kludge_CLI* CLI,
                            bool own_only = false, bool initiate_inheritance = true);
   int pwd(string* result);
   string* My_Name() { return(&Name); }
   Kludge_Object* Superior();

   // originally global functions - various lookups and DB op wrappers
   static Kludge_Class* ID2ptr(OID_Type* ID, Kludge_CLI* CLI);
   static Kludge_Class* Name2ptr(string* Class_Name, Kludge_CLI* CLI);
private:
   static Kludge_MYSQL_RES* DB_Lookup_Internals(char* qry_string, Kludge_CLI* CLI);
   static Kludge_MYSQL_RES* DB_Lookup(string* Class_Name, Kludge_CLI* CLI);
   static Kludge_MYSQL_RES* DB_Lookup(OID_Type Class_ID, Kludge_CLI* CLI);
   static Kludge_Class* Load_By_ID(Kludge_MYSQL_RES* row, Kludge_CLI* CLI);
   static Kludge_Class* Name2ptr_Raw(string* Class_Name);
   static Kludge_Class* ID2ptr_Raw(OID_Type* ID);
   // traversal callbacks, for use with GTrees:
   static int List_Children_Traverse(gpointer key, gpointer Child_Class, gpointer CLI);
   static int List_Children_Names_Traverse(gpointer key, gpointer Child_Class, gpointer CLI);
   static int List_Properties_Traverse(gpointer key, gpointer Property, gpointer CLI);
   static int List_Properties_Names_Traverse(gpointer key, gpointer Property, gpointer CLI);
   static int List_Properties_Inh_Traverse(gpointer key, gpointer Property, gpointer CLI);
   static int List_Properties_Inh_Names_Traverse(gpointer key, gpointer Property, gpointer CLI);
   static int List_Relationships_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI);
   static int List_Relationship_Names_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI);
   static int List_Relationships_Inh_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI);
   static int List_Relationship_Inh_Names_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI);
   static int Free_Properties_Traverse(gpointer key, gpointer Property, gpointer data);
   static int Free_Children_Traverse(gpointer key, gpointer Child_Class, gpointer data);
   static int Free_Relationships_Traverse(gpointer key, gpointer Relationship, gpointer rels_to_free);
   static int ReInherit_Children_Traverse(gpointer key, gpointer Child, gpointer CLI);
   static int UnInherit_Children_Traverse(gpointer key, gpointer Child, gpointer data);
   static int Build_Props_Array_Traverse(gpointer key, gpointer Content, gpointer array_to_build);
   static int Build_Rels_Array_Traverse(gpointer key, gpointer Content, gpointer array_to_build);

   bool Extract_From_MysqlRow(MYSQL_ROW Row_To_Process);
   Kludge_Class_Rel_Stub* Lookup_Relationship_Raw(string* name_to_lookup, Kludge_CLI* CLI);
};



// Corresponding to an entry from table Class_Properties
class Kludge_Class_Property : public Kludge_Object
{
public:
   Kludge_Class* Class; // needed for zapping and for the write-back mechanism
                        // Othewise the property is linked to from the owning
                        // class by the "Properties" member of class Kludge_Class
   string RDN;
   string Value;
   bool Suppress;       // not used yet...
   bool ctor_failed;

#define CLSPROP_IS_BLOB     1
#define CLSPROP_IS_TXT_BLOB 2
#define CLSPROP_IS_METHOD   4
   // A BLOB is a leaf object, just like a property - except for the potentially huge
   // volume of data.
   // I was intending to implement BLOBs properly using C++ inheritance, as a child class
   // of Kludge_Class_Property - it's just that all the properties are stored in
   // a single relational table, so I need Kludge_Class_Property::Extract_From_MysqlRow()
   // to already extract BLOBs as well...
   enum Property_SubType    // For RTTI 
   {
      SUBTYPE_PROPERTY,
      SUBTYPE_BLOB,
      SUBTYPE_TXT_BLOB,
      SUBTYPE_METHOD
   };
   Property_SubType SubType;
   // The BLOB contents are never kept cached in RAM - they're stored straight
   // into the database upon INSERT/SET and they're only loaded (and shoveled
   // straight to the client) upon GET. Even that way, due to the transcodings
   // necessary, we can't avoid having two or three copies on the heap while
   // the transaction is being performed. Subject to further discussion.

   // the following two methods apply only to SUBTYPE_METHOD
   int UnParse_Method();
   int Parse_Method(Kludge_CLI* CLI);
   int Run_Method(Kludge_CLI* CLI, Kludge_pl_params* args); // deletes the args vector

   Kludge_Class_Property(MYSQL_ROW DB_record, Kludge_Class* master_class, Kludge_CLI* CLI);
   Kludge_Class_Property(string* cls_prop_name, Kludge_Class* master_class,
                         string* val, Kludge_CLI* CLI, bool sup = false,
                         Property_SubType _type = SUBTYPE_PROPERTY, byte_buf* blob_value = NULL);
   ~Kludge_Class_Property();

   // In the following two function headers, "NULL" actually means "this->Class":
   int Add_To_Index(Kludge_Class* Where = NULL);
   int Free_From_Index(Kludge_Class* Where = NULL);
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int Describe_Self(Kludge_CLI* CLI);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI); // does not delete obj_to_insert
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int Delete(Kludge_CLI* CLI, bool recursive = false);
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false);          // DB access function
   int Zap();
   int pwd(string* result);
   string* My_Name() { return(&RDN); }
   Kludge_Object* Superior();
private:
   // The following CV* allows us to keep parsed methods (OP* trees) loaded on the heap.
   CV* Method_CV; // Any method calling SvREFCNT_dec on this must also set this to NULL!
   bool Extract_From_MysqlRow(MYSQL_ROW Row_To_Process);
   int Report_Subtype(Kludge_CLI* CLI);
   int Set_Blob(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   Kludge_MYSQL_RES* Get_Blob_Content(Kludge_CLI* CLI);
};



class Kludge_Class_Relationship; // to allow a mutually recursive declaration

// The following is a helper class to allow Kludge_Class_Relationship to behave
// in a "master" or "slave" way, depending on which way it was resolved from
// the CLI path. E.g., to state either master or slave name in List_Contents().
class Kludge_Class_Rel_Stub : public Kludge_Object
{
public:
   enum Rel_Stub_Type    // To distinguish between Master_Stub and Slave_Stub
   {
      MASTER,
      SLAVE
   };

   Rel_Stub_Type Stub_Type;
   Kludge_Class_Relationship* Body;
   Kludge_Class* Owner;
   string Name;
   unsigned int MinCount;
   unsigned int MaxCount;
   string NamingGen;
   string NamingMatch;
   
   Kludge_Class_Rel_Stub(Rel_Stub_Type stub_type, Kludge_Class_Relationship* relationship_body);
   ~Kludge_Class_Rel_Stub();
   int Zap();
   int Init_Owner(Kludge_CLI* CLI);
   int Add_To_Owner_Index(Kludge_Class* Where = NULL);
   int Free_From_Owner_Index(Kludge_Class* Where = NULL);
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int Describe_Self(Kludge_CLI* CLI);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false);
   int Delete(Kludge_CLI* CLI, bool recursive = false);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int pwd(string* result);
   string* My_Name() { return(&Name); }
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   Kludge_Object* Superior();
   int SetNamingMatch(string* pattern, Kludge_CLI* CLI);
   int Compile_Regex(Kludge_CLI* CLI);
   int Match_Name(string* subject, Kludge_CLI* CLI);

private:
   pcre* NamingMatchCompiled;
};



// Corresponding to an entry from table Class_Relationships
//class Kludge_Class_Relationship : public Kludge_Object
class Kludge_Class_Relationship
{
public:
   Kludge_Class_Rel_Stub Master_Stub;
   Kludge_Class_Rel_Stub Slave_Stub;
   OID_Type RelCLSID;
   int RelType; // 0 = peer2peer, 1 = master/slave, 2 = ownership 
   unsigned int DefaultCount_M;
   // The following static member points to a global index of Kludge_Class_Relationship objects.
   // It's initialized at the top of kludge_engine.cc (mandatory static member init).
   static GTree* Index;
   GTree* Instances;  // for instance relationships (and owned instances and properties) - see RelTYpe
   bool Suppress;
   bool ctor_failed;

   Kludge_Class_Relationship();
   int Init(Kludge_CLI* CLI); // original heavy-weight constructor
   Kludge_Class_Relationship(MYSQL_ROW DB_record, Kludge_CLI* CLI);
   ~Kludge_Class_Relationship();
   int Zap();
   bool Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI);
   int Init_Instances();
   bool Free_Instances();
   int Add_To_Index();
   int Free_From_Index();
   int Add_To_Global_Index();
   int Free_From_Global_Index();
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int DB_Delete(Kludge_CLI* CLI);
   int Delete(Kludge_CLI* CLI);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI);
   int pwd(string* result);
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);

   static Kludge_Class_Relationship* ID2ptr(OID_Type* ID, Kludge_CLI* CLI);
private:
   static Kludge_Class_Relationship* ID2ptr_Raw(OID_Type* ID);
   static Kludge_MYSQL_RES* DB_Lookup(OID_Type rel_clsid, Kludge_CLI* CLI);
};



// for a mutually recursive declaration :
class Kludge_Instance_Property;
class Kludge_Inst_Rel_Stub;

// Corresponding to an entry from table Instances
class Kludge_Instance : public Kludge_Object
{
public:
   OID_Type OID;
   Kludge_Class* Class;
   Kludge_Class_Relationship* ClsRel;
   string RDN;
   Kludge_Instance* Owner;
   // The following static member points to a global index of Kludge_Instance objects.
   // It's initialized at the top of kludge_engine.cc (mandatory static member init).
   static GTree* Index;
   GTree* Instances_Owned;
   GTree* Properties;
   GTree* Relationships;      // if NULL, relationships index not initialized
   bool relationships_loaded;
   bool ctor_failed;

   Kludge_Instance(string* instance_name, Kludge_Instance* owner_instance,
                 Kludge_Class* inst_class, Kludge_Class_Relationship* cls_rel, Kludge_CLI* CLI);
   Kludge_Instance(const char *instance_name, Kludge_Instance* owner_instance, Kludge_CLI* CLI);
   Kludge_Instance(MYSQL_ROW DB_record, Kludge_Instance* owner_instance, Kludge_CLI* CLI);
   ~Kludge_Instance();
   bool Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI);
   int Load_Instances_Owned(Kludge_CLI* CLI, bool recursive = false);
   bool Free_Instances_Owned();
   int Load_Properties(Kludge_CLI* CLI);
   bool Free_Properties();
   int Init_Relationships();
   int Load_Relationships(Kludge_CLI* CLI);
   bool Free_Relationships();
   int Add_To_Owner_Index(Kludge_Instance* Where = NULL);
   int Add_To_Class_Index(Kludge_Class* Where = NULL);
   int Add_To_ClsRel_Index(Kludge_Class_Relationship* Where = NULL);
   int Add_To_Index(); // all the three above, plus global
   int Free_From_Owner_Index(Kludge_Instance* Where = NULL);
   int Free_From_Class_Index(Kludge_Class* Where = NULL);
   int Free_From_ClsRel_Index(Kludge_Class_Relationship* Where = NULL);
   bool Free_From_Global_Index();
   int List_Properties(Kludge_CLI* CLI, bool recursive = false);
   int List_Instances_Owned(Kludge_CLI* CLI, bool recursive = false);
   int List_Relationships(Kludge_CLI* CLI, bool recursive = false);
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int Describe_Self(Kludge_CLI* CLI);
   Kludge_Instance* Lookup_Instance_Owned(string* name_to_lookup, Kludge_CLI* CLI);
   Kludge_Instance_Property* Lookup_Property(string* name_to_lookup, Kludge_CLI* CLI);
   Kludge_Inst_Rel_Stub* Lookup_Relationship_Raw(string* name_to_lookup, Kludge_CLI* CLI);
   Kludge_Inst_Rel_Stub* Lookup_Relationship(string* name_to_lookup, Kludge_CLI* CLI);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI); // does not delete obj_to_insert
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int Delete(Kludge_CLI* CLI, bool recursive = false);
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false);          // DB access function
   int Zap();
   int pwd(string* result);
   string* My_Name() { return(&RDN); }
   Kludge_Object* Superior();
   int Try_Method(char* method_name, Kludge_pl_params* args,    // Try_Method() adds "this" to args.
                  Kludge_CLI* CLI, bool own_only = false);      // If args == NULL, they're created.
   
   // originally global functions - various lookups and DB op wrappers
   static Kludge_Instance* ID2ptr(OID_Type* ID, Kludge_CLI* CLI);
private:
   static Kludge_MYSQL_RES* DB_Lookup(OID_Type OID, Kludge_CLI* CLI);
   static Kludge_Instance* Load_By_ID(Kludge_MYSQL_RES* row, Kludge_CLI* CLI);
   static Kludge_Instance* ID2ptr_Raw(OID_Type* ID);
   // traversal callbacks
   static int List_Properties_Traverse(gpointer key, gpointer Property, gpointer CLI);
   static int List_Properties_Names_Traverse(gpointer key, gpointer Property, gpointer CLI);
   static int List_Instances_Owned_Traverse(gpointer key, gpointer Instance_Owned, gpointer CLI);
   static int List_Instances_Owned_Names_Traverse(gpointer key, gpointer Instance_Owned, gpointer CLI);
   static int List_Relationships_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI);
   static int List_Relationship_Names_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI);
   static int Free_Properties_Traverse(gpointer key, gpointer Property, gpointer data);
   static int Free_Instances_Owned_Traverse(gpointer key, gpointer Owned_Instance, gpointer data);
   static int Free_Relationships_Traverse(gpointer key, gpointer Relationship, gpointer rels_to_free);
};



// Corresponding to an entry from table Instance_Properties
class Kludge_Instance_Property : public Kludge_Object
{
public:
   OID_Type OID;  // Added only so that we have something for a key with
                  // Kludge_Class::Instances and Kludge_Class_Relationship::Instances.
                  // Perhaps a runtime-only self pointer would do the same job.
   Kludge_Instance* Instance;
   Kludge_Class* Class;
   Kludge_Class_Relationship* ClsRel;
   string RDN;
   string Value;
   bool ctor_failed;

#define INSTPROP_IS_BLOB     1
#define INSTPROP_IS_TXT_BLOB 2
   // A BLOB is a leaf object, just like a property - except for the potentially huge
   // volume of data.
   // I was intending to implement BLOBs properly using C++ inheritance, as a child class
   // of Kludge_Instance_Property - it's just that all the properties are stored in
   // a single relational table, so I need Kludge_Instance_Property::Extract_From_MysqlRow()
   // to already extract BLOBs as well...
   enum Property_SubType    // For RTTI 
   {
      SUBTYPE_PROPERTY,
      SUBTYPE_BLOB,
      SUBTYPE_TXT_BLOB,
   };
   Property_SubType SubType;
   // The BLOB contents are never kept cached in RAM - they're stored straight
   // into the database upon INSERT/SET and they're only loaded (and shoveled
   // straight to the client) upon GET. Even that way, due to the transcodings
   // necessary, we can't avoid having two or three copies on the heap while
   // the transaction is being performed. Subject to further discussion.

   Kludge_Instance_Property(string* inst_prop_name, Kludge_Instance* owner_instance,
                  Kludge_Class* inst_prop_class, Kludge_Class_Relationship* cls_rel,
                  string* val, Kludge_CLI* CLI,
                  Property_SubType _type = SUBTYPE_PROPERTY,
                  byte_buf* blob_value = NULL );
   Kludge_Instance_Property(MYSQL_ROW DB_record, Kludge_Instance* owner_instance, Kludge_CLI* CLI);
   ~Kludge_Instance_Property();
   bool Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI); 
   int Add_To_Owner_Index(Kludge_Instance* Where = NULL);
   int Add_To_Class_Index(Kludge_Class* Where = NULL);
   int Add_To_ClsRel_Index(Kludge_Class_Relationship* Where = NULL);
   int Free_From_Owner_Index(Kludge_Instance* Where = NULL);
   int Free_From_Class_Index(Kludge_Class* Where = NULL);
   int Free_From_ClsRel_Index(Kludge_Class_Relationship* Where = NULL);
   int Add_To_Index();
   int Zap();
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int Describe_Self(Kludge_CLI* CLI);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI);
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false);
   int Delete(Kludge_CLI* CLI, bool recursive = false);
   int pwd(string* result);
   string* My_Name() { return(&RDN); }
   Kludge_Object* Superior();
private:
   int Set_Blob(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   int Report_Subtype(Kludge_CLI* CLI);
   Kludge_MYSQL_RES* Get_Blob_Content(Kludge_CLI* CLI);
};



class Kludge_Instance_Relationship; // to allow a mutually recursive declaration

// The following is a helper class to allow Kludge_Instance_Relationship to behave
// in a "master" or "slave" way, depending on which way it was resolved from
// the CLI path. E.g., to state either master or slave name in List_Contents().
class Kludge_Inst_Rel_Stub : public Kludge_Object
{
public:
   enum Rel_Stub_Type    // To distinguish between Master_Stub and Slave_Stub
   {
      MASTER,
      SLAVE
   };

   Rel_Stub_Type Stub_Type;
   Kludge_Instance_Relationship* Body;
   Kludge_Instance* Owner;
   string Name;

   Kludge_Inst_Rel_Stub(Rel_Stub_Type stub_type, Kludge_Instance_Relationship* rel_body);
   ~Kludge_Inst_Rel_Stub(){}
   int Add_To_Owner_Index(Kludge_Instance* Where = NULL);
   int Free_From_Owner_Index(Kludge_Instance* Where = NULL);
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int Describe_Self(Kludge_CLI* CLI);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false);
   int Delete(Kludge_CLI* CLI, bool recursive = false);
   int Zap();
   int Init_Owner(Kludge_CLI* CLI);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int pwd(string* result);
   string* My_Name() { return(&Name); }
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   Kludge_Object* Superior();
};



// Corresponding to an entry from table Instance_Relationships
//class Kludge_Instance_Relationship : public Kludge_Object
class Kludge_Instance_Relationship
{
public:
   Kludge_Inst_Rel_Stub Master_Stub;
   Kludge_Inst_Rel_Stub Slave_Stub;
   OID_Type RelOID;
   Kludge_Class_Relationship* ClsRel;
   bool ctor_failed;

   Kludge_Instance_Relationship();
   int Init(Kludge_CLI* CLI); // original heavy-weight constructor
   Kludge_Instance_Relationship(MYSQL_ROW DB_record, Kludge_CLI* CLI);
   ~Kludge_Instance_Relationship();
   int Zap();
   bool Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI);
   int Add_To_ClsRel_Index(Kludge_Class_Relationship* Where = NULL);
   int Add_To_Index();
   int Free_From_ClsRel_Index(Kludge_Class_Relationship* Where = NULL);
   int Free_From_Index();
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   int DB_Delete(Kludge_CLI* CLI);
   int Delete(Kludge_CLI* CLI);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI);
   int pwd(string* result);
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
};



class Kludge_Root_Object : public Kludge_Object
{
public:

   Kludge_Class* Classes;       // the Root Class
   Kludge_Instance* Instances;  // the Root Instance
   string Name;

   Kludge_Root_Object();
   ~Kludge_Root_Object();
   int List_Contents(Kludge_CLI* CLI, bool recursive = false);
   Kludge_Object* Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false);
   int Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI); // does not delete obj_to_insert
   int Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI);
   int Move(Kludge_Object* destination, Kludge_CLI* CLI);
   int Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI);
   int Delete(Kludge_CLI* CLI, bool recursive = false);
   int DB_Delete(Kludge_CLI* CLI, bool recursive = false);          // DB access function
   int Zap();
   int pwd(string* result);
   string* My_Name() { return(&Name); }
   Kludge_Object* Superior();
};


// It's probably a bad habit to put extern variable declarations
// into a header file, but this one's worked for me so far.
extern Kludge_Root_Object* Kludge_Root;







/* ###    global functions (stand-alone, non-member)    ### */

// GTree traversal and comparison functions

int Kludge_Delete_Contents_Traverse(gpointer key, gpointer Content, gpointer del_parm_passed);

int Kludge_Compare_Names(gconstpointer a, gconstpointer b); // string comparison
int Kludge_Compare_IDs(gconstpointer a, gconstpointer b);   // integer comparison

// a cleanup function for the global indexes (declared and defined static)
int Clean_Up_Global_Indexes();

#endif // _KLUDGE_ENGINE_H

