/*

This is              kludge_perl.h

This module contains glue code between kludge and Perl.
More glue code is in the accompanying xs addition file,
Kludge_pl_xs.xs

*/

#ifndef _KLUDGE_PERL_H
#define _KLUDGE_PERL_H

#include <string>
#include <vector>

#undef MIN
#undef MAX
#include <EXTERN.h>
#include <perl.h>
#undef Move
#undef assert

class Kludge_Object;  // another forward declaration to alleviate mutually recursive inclusion

// The class and typedef below are used to pass parameters in a uniform way
// from the C++ code into Perl subroutines.
//
// We're passing strictly an array of named parameters.
// The Perl-space kludge methods MUST create an anonymous hash as their
// first operation, and include the obtained array in it. This is where
// default values can be inserted. (source: the Perl Cookbook)
//
// Example: Dog::Bark()
// 
//    #sub {                          # this line is added automatically on the fly
//       my %args = (                 # mandatory lead-in
//           INTENSITY => 'medium',   # optional line - default value of an argument
//           PITCH     => 'high',     # optional line - default value of an argument
//           @_                       # this line loads the argument list passed - mandatory
//       );                           # end of mandatory lead-in
//
//       if ( $args{PITCH}  == 'low' )
//       {
//          # more code here...
//       }
//       # more code here...
//    #}                              # this line is added automatically on the fly
//
// Likewise, Perl-space subs calling other kludge methods MUST pass
// parameters in this same fashion.
// This convention facilitates direct calls of kludge methods by their CV
// reference, both from C++ and from within Perl space.
//
// At the level of perl guts, the passed array of named parameters is not
// an anonymous hash (which would seem to be the obvious implementation)
// - rather, it's an array of ( key, value, key, value, key, value, ... etc.).
// The elements are pushed onto the perl stack individually, one by one, 
// just before the sub is perl_call()'ed. Only after @_ is included in
// a local anonymous hash (at the start of the sub being called), the elements
// become true key=>value pairs in a perl hash table.

class Kludge_pl_param
{
public:
   Kludge_pl_param(char* key, char* value)      // the contents are copied - passed by value
     : Key(newSVpv(key,0)),
       Value(newSVpv(value,0))
   {}

   Kludge_pl_param(char* key, string& value)    // the contents are copied - passed by value
     : Key(newSVpv(key,0)),
       Value(newSVpv((char*)value.c_str(),0))   // the typecast just discards const
   {}                                           // (zaps a warning)

   Kludge_pl_param(char* key, int value)        // passed by value
     : Key(newSVpv(key,0)),
       Value(newSViv(value))
   {}

   Kludge_pl_param(char* key, Kludge_Object* value)  // passed by reference
     : Key(newSVpv(key,0)),
       Value(newSViv((int)value))                    // the pointer gets stuffed into an int
   {}                                                //  ==> no copy construction occurs

   Kludge_pl_param(char* key, CV* value)       // passed by reference (newRV_inc(CV*))
     : Key(newSVpv(key,0)),                    //  - no copy construction occurs
       Value(newRV_inc((SV*)value))  // Perl uses a rudimentary pre-C++ polymorphism trick.
   {}                                // SV*, CV* et al are declared the same way, to be 
                                     // low-level identical. That's why the typecast works.

   ~Kludge_pl_param(){};

   SV* Key;
   SV* Value;
};

typedef vector<Kludge_pl_param*> Kludge_pl_params;





// The following functions are meant to facilitate inserting (evaluating)
// and running of perl-space subroutines from the C++ space of kludge.

// Evaluate a generic chunk of Perl code in the existing initialized interpreter.
// This is useful for bootstrapping.
int Kludge_pl_eval(char *code_str);

// Call a Perl-space subroutine by name, passing it a single integer parameter.
//int Kludge_pl_call(char *subname, int int_passed);
// No use for this function at the moment. In the future, we may need to call
// some Perl stuff by name from C++ behind the scenes...



// The following function evaluates a passed chunk of Perl code as an
// anonymous sub, in the existing initialized interpreter.
// Pass a bare sub body - i.e., do not include the sub definition grammar 
// (sub <name> { }). The sub definition paraphernalia is added behind the scenes,
// on the fly.
// Example: the definition of Dog::Bark()
// 
//    #sub {                          # this line is added automatically on the fly
//       print "baw, waw\n"           # your code here - this is the only line of user code
//    #}                              # this line is added automatically on the fly
//
CV* Kludge_pl_eval_sub(char* code_str);
CV* Kludge_pl_eval_sub(string& code_str);  // A wrapper taking string& instead of char* .
                                           // Couldn't be inlined - the linker kept failing :(

// The following function calls a sub by CV (useful for the anonymous
// subs of Kludge). It does delete the params array passed and its member
// structs - don't delete these in the calling code.
// If the args pointer is null, this is evaluated as no parameters
// (still a valid call).
// Results are discarded.
int Kludge_pl_call_CV(CV* sub_CV, Kludge_pl_params* args);



// The following four functions mimick the original thread creation
// stuff from Thread.c.
// The threadstart...() sisters need to be called upon startup and
// shutdown of a child thread, from within the child thread.
// The newthread...() sisters need to be called from within the
// parent thread, pthread_create()ing the children - before and
// after pthread_create().

// We need to pass the current Perl Thread as an argument to threadstart_LeadIn()
// and threadstart_LeadOut(). The thread is created by newthread_LeadIn - you
// need to pass it into the child thread as part of your custom per-thread
// struct*, that's given as the last parameter to pthread_create().
// Or in some other way - that's up to you, the user of this library.
int threadstart_LeadIn(Thread arg);
int threadstart_LeadOut(Thread arg);

// We need to pass thr, oldmask and pthread_create() result code to the leadout.
// Please declare n_thread, oldmask and thr_perl_err in the calling scope.
// Like this:
//
// Thread n_thr;       // threaded perl stuff - gee, what about using C++ references?
// sigset_t oldmask;   // threaded perl stuff
// int thr_perl_err;   // threaded perl stuff
//
// thr_perl_err = newthread_LeadIn(&n_thr, &oldmask);
// pthread_create( some attributes my C++ code is using );
// newthread_LeadOut(thr_perl_err, &n_thr, &oldmask);
int newthread_LeadIn(Thread* n_thread, sigset_t* oldmask);
int newthread_LeadOut(int err, Thread* n_thread, sigset_t* oldmask);



// The following two  functions need to be called within main() - one
// at startup, the other one at shutdown.

// Create the single global interpreter, parse the null/bogus script,
// prepend our private perl dir into the @INC path, load the necessary
// modules (Thread, Kludge_pl, DynaLoader)
int Kludge_pl_init();

// Tear down and deallocate the global Perl interpreter.
void Kludge_pl_shutdown();

// A helper - we also need to know the Thread struct of the default thread.
Thread Kludge_pl_get_deflt_thread();



// The following function needs to be called within the client thread
// to init and tear down some Kludge-wise thread-private variables.
// In other words, these initializations are not a part of the original
// threadstart() function - rather, they're some Kludge-proprietary Perl-space
// initial magic and "work behind the scenes" to help the Perl-space "user"
// programmer.
// This function is a wrapper around the once-standalone Kludge_CLI::Run()
int Kludge_pl_thr_Run();



// A simple wrapper around SvTYPE - reports the SV type as a character string.
string* decipher_SV_type(SV* data);

#endif // _KLUDGE_PERL_H
