/* 
-- A rudimentary multi-threaded TCP sockets server, 
-- listening on a single interface,
-- only splinting off a child thread upon the arrival of a new connection.
-- Plus some supportive functions, such as the GNU getopt() handling.
*/

#include <string>
#include "kludge_threads.h" // also includes <fstream>, <pthread.h>
#include <string.h>      // provides strerror(), memcpy()
#include <sys/types.h>   // provides some pthread.h and socket.h types
#include <sys/socket.h>  // provides socket(), bind(), listen(), accept()
#include <netdb.h>       // provides gethostbyname()
#include <netinet/in.h>  // provides htonl(), sockaddr_in

#define _GNU_SOURCE
#include <getopt.h>      // provides getopt_long()      (only used for --help)


int fatal_error;
// a global variable for fatal error signaling to the final exit();

bool shutdown_initiated, shutdown_forced;
Kludge_Thread* Pseudo_Master_Thread; // contains the default mysql connection
pthread_t master_thread;             // used by Kludge_CLI.cc to send a shutdown signal

// mandatory definition of a static member:
GTree* Kludge_Thread::threads = g_tree_new(Kludge_Compare_IDs);






// ###    class Kludge_Thread    ###


Kludge_Thread::Kludge_Thread(int sock, Thread perl_thr)
 : DB(NULL), pl_thr(perl_thr), socket(sock), socket_glob(NULL)
{
   /*(iostream*)*/ input_stream = new fstream(socket);
   /*(iostream*)*/ output_stream = new fstream(socket);
   CLI = new Kludge_CLI(input_stream, output_stream, this);
   pthread_attr_init(&attributes);
   pthread_attr_setdetachstate(&attributes, PTHREAD_CREATE_DETACHED);
}



Kludge_Thread::~Kludge_Thread()
{
   if (input_stream != NULL)
   { 
      input_stream->clear();
      if (socket != 1)
      {
         ((fstream*)input_stream)->close(); // we shouldn't close stdout (FD 1)
      }
      delete input_stream;
   }

   if (output_stream != NULL)
   { 
      *output_stream << flush;
      output_stream->clear();
      if (socket != 1)
      {
         ((fstream*)output_stream)->close(); // we shouldn't close stdout (FD 1)
      }
      delete output_stream;
   }
   
   if (CLI != NULL)
   {
      delete CLI;
   }
   
   if (DB != NULL)
   {
      mysql_close(DB);
   }
   
   pthread_attr_destroy(&attributes); // does nothing in LinuxThreads - included anyway
   // Also, it probably doesn't make much sense to keep the struct initialized
   // throughout the life of a thread - it's only needed for the init.
   // Come to think of it, it needn't be thread-private :)
}



MYSQL_RES* Kludge_Thread::DB_Query(const char* query_string)
{
 MYSQL_RES* result = NULL;

   if (DB == NULL) Open_DB();
   if (DB != NULL) result = kludge_mysql_query(query_string, DB);

   return(result);
}  



int Kludge_Thread::DB_Cmd(const char* query_string)
{
 int result = 0;
 
   if (DB == NULL) Open_DB();
   if (DB != NULL) result = kludge_mysql_cmd(query_string, DB);
   else result = 1;
   
   return(result);
}



int Kludge_Thread::Open_DB()
{
 int result = 0;

   if (DB == NULL) DB = kludge_mysql_init();
   if (DB == NULL) result = 1;

   return(result);
}



/*
-- The following is the client thread function.
-- Originally a global function, now a static member of class Kludge_Thread.
-- Takes a void pointer and returns a void pointer.
-- The void pointer argument is really a Kludge_Thread*, pointing to
-- a fresh instance of class Kludge_Thread, associated with a socket to serve.
--
-- This can't become a regular Kludge_Thread member function - we have to
-- pass the thread function pointer to pthread_create();
*/
void *Kludge_Thread::Run(void *this_kludge_thread)
{
 sigset_t old_mask, my_mask;

   cout << "New socket thread started.\n" ;
   // If fast creation of threads becomes a priority, the Kludge_Thread class
   // constructor should be stripped down as much as possible and most of its
   // functions should be performed by this pthread_create'd thread function.
   // Perhaps there could also be a pool of idle threads waiting for a chance.
   // Nevertheless, as this server is session-based rather than stateless,
   // it may never become a priority to launch threads real quick.
   
   // This needs to be done before Perl init.
   ((Kludge_Thread*)this_kludge_thread)->Add_To_Index();    // can't be done in the ctor - no thread ID
   
   // Initialize Perl for this thread
   threadstart_LeadIn( ((Kludge_Thread*)this_kludge_thread)->pl_thr );

   // Set up signal handling for this child thread 
   //  It would seem that Perl sets full mask within newthread_LeadIn...
   //  (and sets it back to original within the leadout).
   //  This is probably to avoid hiccups during pthread_create().
   //  It would certainly be interesting to find out what signals (if any)
   //  the Perl threading model relies on - i.e., what mask should the children have.
   //  My current hypothesis is that the Perl threads don't care about signals inherently.
   sigemptyset(&my_mask);                           // create an empty struct sigset_t
   sigaddset(&my_mask,SIGINT);                      // add SIGINT to the set
   pthread_sigmask(SIG_BLOCK,&my_mask,&old_mask);   // block SIGINT in this thread
   sigemptyset(&my_mask);                           // create an empty struct sigset_t
   sigaddset(&my_mask,SIGQUIT);                     // add SIGQUIT to the set
   pthread_sigmask(SIG_UNBLOCK,&my_mask,&old_mask); // unblock SIGQUIT in this thread
   
   //((Kludge_Thread*)this_kludge_thread)->CLI->Run(); // now wrapped in a Perl-space scope
   Kludge_pl_thr_Run();                             // NOW FINALLY RUN THE CLI :)

   // Hmm. Once I unblock SIGINT, the thread immediately catches it :-/
   // It's accumulated since the moment I sent it to the master thread 
   // (if sent from outside - by kill() rather than pthread_kill() ).
   //pthread_sigmask(SIG_SETMASK,&old_mask,NULL); // restore signal mask

   // Shutdown Perl in this thread
   threadstart_LeadOut( ((Kludge_Thread*)this_kludge_thread)->pl_thr );
   
   cout << "Client thread done.\n" ;
   
   // This should better be done after we've finished shutting down Perl.
   ((Kludge_Thread*)this_kludge_thread)->Free_From_Index(); // we shouldn't better do this in the dtor

   // as long as we don't have the main() thread catch all the children back
   // and delete the respective Kludge_Thread class instances, we should do this
   // within the thread function.
   delete((Kludge_Thread*)this_kludge_thread);

   pthread_exit(NULL);
   // return(NULL);
}



int Kludge_Thread::Add_To_Index()
{
   cout << "Adding thread " << thread_ID << " to the index of client threads.\n";
   g_tree_insert(threads,&thread_ID,this);
   return(0);
}



int Kludge_Thread::Free_From_Index()
{
   cout << "Removing thread " << thread_ID << " from the index of client threads.\n";
   g_tree_remove(threads,&thread_ID);
   return(0);
}



Kludge_Thread* Kludge_Thread::Lookup_By_ID(pthread_t thr_id)
{
   return( (Kludge_Thread*) g_tree_lookup(threads, &thr_id) );
}



int Kludge_Thread::Kill_Em_All()
{
 vector<Kludge_Thread*> threads_to_kill;

   g_tree_traverse(
         threads,
         Kludge_Thread::Kill_Em_All_Traverse,
         G_POST_ORDER,
         &threads_to_kill);

   for (unsigned int i=0;i<threads_to_kill.size();i++)
   {
      pthread_kill(threads_to_kill[i]->thread_ID,SIGQUIT);
   }

   return(0);
}



int Kludge_Thread::Kill_Em_All_Traverse(gpointer key, gpointer thr, gpointer threads_to_kill)
{
   ((vector<Kludge_Thread*>*)threads_to_kill)->push_back((Kludge_Thread*)thr);
   return(false);
}



int Kludge_Thread::Index_Teardown()
{
   g_tree_destroy(Kludge_Thread::threads);
   return(0);
}








// ###    Global functions    ###


/* 
-- Function tcp_listen_at
-- Takes the DNS name (or IP) and port to lurk at.
-- Returns an open socket with bind() and listen() already applied to it.
-- example: descriptor = tcp_listen_at("localhost", 23);
-- when finished with the connection, close the socket using close().
*/
int tcp_listen_at(
                   const char *listen_interface,
                   const unsigned short int listen_port,
                   const int queue_len
                 )
{
 int                  listen_socket;
 struct hostent       *listen_address_stuff;
 struct protoent      *TCP_proto_stuff;
 struct sockaddr_in   listen_address;

   TCP_proto_stuff = (struct protoent *) getprotobyname("tcp");
   listen_socket = socket(AF_INET, SOCK_STREAM, TCP_proto_stuff->p_proto);
   if (listen_socket == -1)     //error detection
   {
      return(-1);
   }

   listen_address_stuff = gethostbyname(listen_interface);
   if (listen_address_stuff == NULL)     //error detection
   {
      return(-1);
   }

   listen_address.sin_family = AF_INET;
   listen_address.sin_port = htons(listen_port);
   memcpy(
          (void *) &listen_address.sin_addr.s_addr, 
          (void *) listen_address_stuff->h_addr_list[0],
          4
         );
   // htonl not needed - the address is already in network order :)

   // On the next two lines, the SO_REUSEADDR is really important.
   // This is to avoid the "can't bind socket: address already in use" error
   // when the server is re-run immediately after it has exit()ted.
   int on = 1;
   setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on)); 

   if ( bind(
             listen_socket, 
             (sockaddr *) &listen_address, 
             sizeof(struct sockaddr_in)
            ) != 0
      ) 
   {
      return(-1);    //error detection
   }

   if (listen(listen_socket, queue_len) != 0)
   {
      return(-1);    //error detection
   }
   else return(listen_socket);
}



/*
-- just a simple function to take care of processing errors 
-- after GetHostByName
*/
int process_error()
{
   fatal_error = errno;
   if (errno == ENOENT)
   {
      cout << "Domain name resolution error: ";
      switch (h_errno)
      {
         case HOST_NOT_FOUND:
            cout << "host not found.\n";
            break;
         case NO_ADDRESS:
            cout << "host name valid, no address assigned.\n";
            break;
         case NO_RECOVERY:
            cout << "DNS resolver failed for no obvious reason.\n";
            break;
         case TRY_AGAIN:
            cout << "temporary DNS resolver outage. Try again later.\n";
            break;
      }
   }
   else cout << "Error: " << strerror(errno) << "\n";

   return(0);
}



/*
-- Accept()s clients and launches the threads to serve them.
*/
int accept_clients(int socket_to_accept_from)
{
 int accepted_client_socket;
 Kludge_Thread       *another_thread = NULL;
 socklen_t           sockaddr_in_len = sizeof(struct sockaddr_in);
 struct sockaddr_in  accepted_client_address;
 struct sigaction sa;                      // for signals
 Thread n_thr;       // threaded perl stuff
 sigset_t oldmask;   // threaded perl stuff
 int thr_perl_err;   // threaded perl stuff

   shutdown_initiated = false; shutdown_forced = false;
   master_thread = pthread_self();
  
   // To allow the child threads to request shutdown, we need to mess with the
   // signal handlers a little. The master thread spends most of its time waiting
   // for new incoming connections in accept(). This system call must be
   // interrupted once shutdown is requested.
   // When I used signal() to replace the signal handlers, the first sigaction()
   // below would read struct sigaction that still featured SIG_DFL instead of my
   // signal handler. Thus, in order to use the struct sigaction, I had to replace
   // that handler reference anyway - hence an inital signal() became superfluous.
   //
   sigaction(SIGINT,NULL,&sa);           // pop the current SIGINT handling attributes
   sa.sa_handler = Sig_Shutdown_Handler; // replace the default handler with my own routine
   sa.sa_flags &= ~ SA_RESTART;          // prevent accept() from continuing when SIGINT arrives
   sigaction(SIGINT,&sa,NULL);           // make the new settings effective
   sa.sa_handler = Sig_Quit_Handler;     // set this up for the child threads
   sigaction(SIGQUIT,&sa,NULL);          // The set handler is effective for all the threads anyway.
   
   while (shutdown_initiated == false)
   {
      cout << "Master: waiting for incoming connections...\n";

      accepted_client_socket = accept(
                                       socket_to_accept_from,
                                       (sockaddr *) &accepted_client_address,
                                       &sockaddr_in_len
                                      );

      if ( accepted_client_socket != -1 ) // if what we've got is a valid socket
      {
         cout << "Master: ...got one!\n";

         newthread_LeadIn(&n_thr, &oldmask); // Perl thread struct creation
         // The only data we fill in right here is the necessary parameters
         // to pthread_create, the socket descriptor to be served by
         // the new thread, and the created Perl thread struct.
         another_thread = new Kludge_Thread(accepted_client_socket, n_thr); // kludge thread object creation
         
         thr_perl_err = pthread_create(
                              &another_thread->thread_ID,
                              &another_thread->attributes,  // initialized in Kludge_Thread::<CTOR>
                              Kludge_Thread::Run,   // the thread function
                              (void *) another_thread
                                      );
         if (thr_perl_err != 0)
         {
            cout << "Error creating a new thread.\n";
            delete another_thread;
         }
         // else thread created successfully
         newthread_LeadOut(thr_perl_err, &n_thr, &oldmask); // Perl thread final polishing
      }
      else if (errno == EINTR)
      {
         cout << "Shutdown requested.\n";
         shutdown_initiated = true;
      }
      else
      {
         cout << "Hmm. Error accepting client connection on the master socket.\n";
         shutdown_initiated = true;
      }
   }
   
   cout << "Waiting for all the child threads to finish...\n"
        << " (there are " << g_tree_nnodes(Kludge_Thread::threads) << " active threads.)\n";

   shutdown_forced = false;
   while (g_tree_nnodes(Kludge_Thread::threads) > 0)
   {
      sleep(1);
      if (shutdown_forced)  // with LinuxThreads, sleep() alone returns unreliable result codes
      {
         cout << "Allright, let's try to cancel whatever they're doing.\n";
         sleep(1); // Without this, the thread that issued shutdown didn't
                   // make it back to the read() call on the socket served
                   // and exited without properly throwing an exception.

         Kludge_Thread::Kill_Em_All();

         break;
      }
   }

   shutdown_forced = false;
   while (g_tree_nnodes(Kludge_Thread::threads) > 0)
   {
      sleep(1);
      if (shutdown_forced)
      {
         cout << "Allright, shutting down immediately.\n";
         break;
         // Now the program terminates without cleaning up the threads,
         // leaving it up to the OS to clean up any undeallocated mess.
      }
   }

   Kludge_Thread::Index_Teardown();

   // reset the SIGINT handling facility to the original state
   sa.sa_handler = SIG_DFL;
   sa.sa_flags |= SA_RESTART;
   sigaction(SIGINT,&sa,NULL);
   sigaction(SIGQUIT,&sa,NULL);

   return(0);
}



int Get_Options(int argc, char** argv,
                string& listen_address,
                unsigned short int& listen_port)
{
 int option_result;
 int result = 0;
 int option_index = 0;
 static struct option long_options[] =
 {
   {"help", 0, 0, 0},
   {0, 0, 0, 0}
 };

   // For a moment I was wondering if it was appropriate to include getopt_long()
   // (a GNU enhancement) as opposed to plain POSIX getopt().
   // I have decided that it was plausible - even for the single long option of --help.
   while ((option_result = getopt_long(argc, argv, "A:L:h:u:p:d:P:",
                                       long_options, &option_index)) != -1)
   {
      switch (option_result)
      {
         case 0 :  // a long option detected
         {
            if (option_index == 0)       // index == 0   ==>   option == "help"
            {
               Shell_Help();
               exit(0);
            }
         }
         break; 
         case 'A' :   // address/hostname to listen at
         {
            listen_address = optarg;
            cout << "Listening at address: " << listen_address << endl;
         }
         break;
         case 'L' :   // local TCP port number to listen at
         {
            istrstream* tmp_str = new istrstream(optarg);
            *tmp_str >> listen_port;
            delete tmp_str; tmp_str = NULL;
            cout << "Listening at port: " << listen_port << endl;
         }
         break;
         case 'h' :   // mysql host name
         {
            mysql_hostname = optarg;
            cout << "MySQL expected at host: " << mysql_hostname << endl;
         }
         break;
         case 'u' :   // mysql username
         {
            mysql_username = optarg;
            cout << "MySQL username: " << mysql_username << endl;
         }
         break;
         case 'p' :   // mysql password
         {
            mysql_password = optarg;
            cout << "MySQL password: " << mysql_password << endl;
         }
         break;
         case 'd' :   // mysql database name/instance
         {
            mysql_db_name = optarg;
            cout << "MySQL DB name: " << mysql_db_name << endl;
         }
         break;
         case 'P' :   // mysql TCP port number
         {
            istrstream* tmp_str = new istrstream(optarg);
            *tmp_str >> mysql_remoteport;
            delete tmp_str; tmp_str = NULL;
            cout << "MySQL expected at port: " << mysql_remoteport << endl;
         }
         break;
         case ':' : 
         case '?' : result = 1;
         break;
         default:
         break;
      }
   }

   if (result != 0)
   {
      cout <<
      "\nObtained some bad command-line options. Valid options are:\n";
      Shell_Help();
   }

   return(result);
}



int Shell_Help()
{
   cout <<
      "\n"
      "  OPTION             DESCRIPTION                                 DEFAULT    \n"
      " -----------------   ------------------------------------------- -----------\n"
      " -A <host || IP>     = local IP address or hostname to listen at \"localhost\"\n"
      " -L <port>           = local TCP port number to listen at         11111     \n"
      " -h <MySQL host>     = MySQL host name or IP address             \"localhost\"\n"
      " -u <MySQL user>     = MySQL username                            \"kludge\"   \n"
      " -p <MySQL passwd>   = MySQL password                            <empty>    \n"
      " -d <MySQL DBname>   = MySQL database/instance name              \"kludge\"   \n"
      " -P <MySQL port>     = MySQL remote TCP port number              <default>* \n"
      " --help              = print this table                          N/A        \n"
      "\n";

   return(0);
}



// ###    Signal handlers    ###

void Sig_Shutdown_Handler(int sig_num)
{
   cout << "SIGINT caught.\n" << flush;
   shutdown_forced = true;
}



void Sig_Quit_Handler(int sig_num)
{
   int this_thread_ID = pthread_self();
   Kludge_Thread* this_thread = (Kludge_Thread*) g_tree_lookup(Kludge_Thread::threads,&this_thread_ID);
   this_thread->CLI->CLI_done = true;
   this_thread->CLI->tokenizer->done = true;
   
   cout << "SIGQUIT caught in thread " << this_thread_ID << " .\n" << flush;
}

