/*

This is              kludge_threads.h

This module contains a few wrapper functions and classes working with
sockets and threads. It takes care of the listening on a socket and
splinting off a thread for each incoming connection. It also plugs the
(int) descriptor of an accepted socket (connection) into two C++ stream
objects, which are later passed to the CLI (as part of an instance of
class Kludge_Thread). It also calls some per-thread Perl init.

This module is called from the main() function to start accepting 
clients, and invokes a command line interface for each client 
connection.

Note: don't confuse the explicit master thread of Kludge with the
hidden "manager" thread, which is an implementation-specific artifact
of POSIX threads in Linux.


    ###   A more detailed note on sockets   ###

The master thread of Kludge is waiting in the accept() call for
incoming TCP sessions.
Whenever a client session comes, the accept() call returns with an
open socket descriptor (int), that is passed to a newly started
child/client thread for further processing. The master thread loops
to call accept() again, the child thread just created starts a CLI
for the particular client.

In terms of per-client-session TCP I/O handling, in the beginning
there's an open (int) BSD socket. As Kludge is written in
(quite kludgey) C++, it prefers to talk to the client via C++ streams,
rather than sockets. The embedded Perl snippets then use a Perl guts
object known as `struct IO'. To sum up, at runtime, a client thread
in Kludge has the following structs hooked up to the original single
socket:

(int) socket
  |
  o--o-->-- (fstream) Kludge_Thread::input_stream
  |  +--<-- (fstream) Kludge_Thread::output_stream
  |
  o-------- (struct IO) contained in (struct GV*) Kludge_Thread::socket_glob

Both C++ fstreams and Perl's struct IO are really using the `struct FILE'
from standard C's stdio.h, which provides buffered IO operations. In this
context, the term "buffered" means that a write() on the socket only occurs
whenever the buffered FILE handle is _flushed_. Apparently, a TCP packet is
generated upon every write() on the socket.

Flushing can occur automatically - at a particular 'high watermark' of the
buffer, after some time, upon a special event (fclose() requested) or after
every fwrite (autoflush can be toggled ON in Perl). A flush can also be
requested explicitly - the underlying call is <stdio.h>'s fflush(FILE*),
which is used by C++ `<< flush' stream manipulator and which seems to be
missing from Perl's file handles API. Now the flush action is available
with Perl's Socket objects in the Socket module - it's just that Kludge
is not using the Socket module. Rather, the socket is hooked up straight
into a file handle called $CLIENT_SOCKET that is then visible from Perl
space. For me, this was the only way (or perhaps the most straight-forward
way) of plugging the bare BSD socket into Perl. Hence, I also had to
write my own fflush XSUB - currently called Kludge_pl::flush_file().

When in comes to using `struct FILE' tied to a bidirectional BSD socket,
sooner or later you have to face strange unrecoverable IO errors when you
attempt mixed reads and writes on the buffered `struct FILE'. Some man
pages and other sources seem to suggest that it should be enough to
interleave subsequent reads and writes with seeks (even null relative
seeks) - but this did not seem to work for me. When there was pending
data on input of a FILE and I wrote something to the same FILE for
output, the thing broke down. (More precisely, I was experiencing this
issue with a single C++ fstream.)
For me, the only pure solution was to create two separate instances of 
`struct FILE' over a single socket: one for reading and one for writing.
Then it doesn't matter if you read all data in the input buffer before
flushing the output. No intermediate seeks are necessary.

Okay - but we're not working with `struct FILE *', right? We're working
with C++ fstream and Perl's `struct IO' and `struct GV' (and a couple
of other IO-related perl guts). So how does the FILE stuff apply?

To be absolutely precise, see the code in Kludge_Thread::Kludge_Thread()
and in create_thr_socket_handle().
There are two distinct C++ fstreams. In GNU C++ with its libg++2,
a fstream is impemented using struct FILE, which is something we
can't see, being mere users of the standard C++ IO library. These
fstreams are constructed in a somewhat non-standard fashion,
with the socket being passed as a single argument to the
constructor.
On the Perl side, there's a single struct IO, pointed to by the GV*
that's a member of class Kludge_Thread. This single struct IO contains
two FILEs, one for input and one for output - Mr. Larry Wall has paved
the way for me here. I'm using some simple hooks in the Perl guts to
fopen the two FILEs over the same BSD socket. The result is a Perl-space
file handle that is really a socket - although there's no such thing
in vanilla Perl, where handles are handles and sockets are sockets :)

To sum up, in the end there are four buffered FILE's open over the
single accept()ed socket.
The hooks between the C++ space and Perl space file IO paraphernalia
and the bare BSD socket may seem to be dirty hacks, but then again
this is Kludge, isn't it? :)
The obvious downside is the implied lack of straight portability :(

Both C++ code and Perl code can read input and write output
almost independently. It's just a matter of flushing the output
explicitly when a flush is due.

*/

#include <fstream>
#include <pthread.h>
#include <signal.h>      // provides the signal handling apparatus
#include <errno.h>       // provides errno
#include <unistd.h>      // provides sleep(), close()
#include "kludge_CLI.h"
#include "kludge_engine.h"
#include <exception>
#include "kludge_perl.h"

#ifndef _KLUDGE_THREADS_H
#define _KLUDGE_THREADS_H



// classes

class Kludge_Thread 
{
public:
   pthread_attr_t attributes;    // some pthread.h thread attributes
   pthread_t      thread_ID;     // thread ID - in Linux probably equal to PID
   Kludge_CLI*    CLI;
   MYSQL*         DB;
   Thread         pl_thr;        // Perl's per-thread data; pl_thr->tid = one-based incremental ID
   static GTree*  threads;

   Kludge_Thread(int sock, Thread perl_thr);
   ~Kludge_Thread();
   MYSQL_RES*     DB_Query(const char* query_string); // returns table entries
   int            DB_Cmd(const char* query_string);   // returns success/failure
   int            Open_DB();                          // returns success/failure
   static void*   Run(void *this_kludge_thread); // an argument for pthread_create()
   int            Add_To_Index();
   int            Free_From_Index();
   static Kludge_Thread* Lookup_By_ID(pthread_t thr_id);
   static int     Kill_Em_All();
   static int     Index_Teardown();
   int            socket;         // BSD socket - a file-handle style descriptor
   GV*            socket_glob;    // a Perl "glob" containing the IO handle to the socket
private:                          //  -- embedded in the struct IO ^^ there are two (struct FILE)'s.
   iostream       *input_stream;  // Two C++ streams tied to the single socket.
   iostream       *output_stream; // A single stream wouldn't do - there were in/out syncing errors.
   static int     Kill_Em_All_Traverse(gpointer key, gpointer thr, gpointer threads_to_kill);
};


// It's probably a bad habit to put extern variable declarations
// into a header file, but this one's worked for me so far.
extern Kludge_Thread* Pseudo_Master_Thread;



// global functions

int tcp_listen_at(
                   const char *listen_interface,
                   const unsigned short int listen_port,
                   const int queue_len
                 );
int process_error();
int accept_clients(int socket_to_accept_from);

int Get_Options(int argc, char** argv,
                string& listen_address,
                unsigned short int& listen_port);
int Shell_Help();


// signals handlers and an exception class
void Sig_Shutdown_Handler(int sig_num);
void Sig_Quit_Handler(int sig_num);
class Ex_Shutdown {};

#endif // _KLUDGE_THREADS_H
